name = "PhyAAt: Physiology of Auditory Attention"
url  = 'https://PhyAAt.github.io'
description = "Predictive Analysis of Auditory Attention from Physiological Signals"

__version__ = '0.0.3'
__author__ = 'Nikesh Bajaj'

from .ProcessingLib import *
from .dataset import *
from .ICA_methods import ICA
from .artifact_correction import *
from . import utils
#import .dataset
